#include "IO.h"

extern GRAPH G;

//-------------------------------------------------------------------------------------
void prettyPrintSolution(SOL &S, ostream &outStream) {
	//Writes a solution in a presentable way to the specified output stream
	int i, j, k = S.items.size(), cnt = 0;
	outStream << "\n\n";
	for (i = 0; i < k; i++) {
		outStream << "Col" << setw(4) << i + 1 << " = { ";
		for (j = 0; j < S.items[i].size(); j++) {
			if (G.precol[S.items[i][j]] != -1) {
				outStream << "[";
				if (!S.isHappy[S.items[i][j]])
					outStream << "#" << S.items[i][j] + 1;
				else 
					outStream << S.items[i][j] + 1;
				outStream << "] ";
			}
			else {
				if (!S.isHappy[S.items[i][j]])
					outStream << "#" << S.items[i][j] + 1 << " ";
				else
					outStream << S.items[i][j] + 1 << " ";
			}
		}
		outStream << "} \n";
		cnt += S.items[i].size();
	}
	outStream << "|V| = " << cnt << ", k = " << G.k << ", numHappy = " << S.numHappy << "\n\n";
}

void printRunDetails(SOL &S, double duration, int UB, ostream &outStream) {
	//Write run details to the specified output stream
	outStream 
		<< "Num vertices      = " << G.n << "\n"
		<< "k                 = " << G.k << "\n"
		<< "Min Degree        = " << G.minDeg << "\n"
		<< "Max degree        = " << G.maxDeg << "\n"
		<< "Total time taken  = " << duration << " sec.\n"
		<< "Happy vertices    = " << S.numHappy << "\n"
		<< "Unhappy vertices  = " << G.n - S.numHappy << "\n"
		<< "Lower bound       = " << S.numHappy << "\n"
		<< "Upper bound       = " << UB << "\n"
		<< "Gap               = ";
	if (UB != 0) outStream << (UB - S.numHappy) / double(UB) << endl;
	else outStream << "0.0" << endl;
	outStream << "\nHere is the solution. []'s indicate precolourings, #'s indicate unhappy vertices:";
	prettyPrintSolution(S, outStream);
}

void printSlnToFile(string filename, SOL &S) {
	//Write a solution to the specidied output file
	ofstream outStream;
	int i;
	string outfilename = "sln_" + filename;
	outStream.open(outfilename);
	if (outStream.fail()) {
		exitAndLog("Error opening output sln file. Exiting...\n");
	}
	outStream << S.numHappy << endl;
	for (i = 0; i < G.n; i++) {
		outStream << S.col[i] + 1 << endl;
	}
	outStream.close();
}

void printRunDetailsToFileAndConsole(string filename, SOL &S, double duration, int UB) {
	ofstream outStream;
	string outfilename = "out_" + filename;
	outStream.open(outfilename);
	if (outStream.fail()) {
		exitAndLog("Error opening output out file. Exiting...\n");
	}
	printRunDetails(S, duration, UB, outStream);
	printRunDetails(S, duration, UB, cout);
	outStream.close();
}

void readInputFile(string filename) {
	//Reads a DIMACS format file and creates the corresponding degree array and adjacency matrix
	ifstream inStream;
	inStream.open(filename);
	if (inStream.fail()) {
		exitAndLog("Error opening input file. Exiting...\n"); 
	}
	char c, str[1000];
	int line = 0, i, j, numEdges = 0, numNodes = 0, numCols = 0, node, colour, edgeCnt = 0;
	vector<vector<bool> > adjacent;
	vector<int> precol;
	vector<bool> colUsed;
	try {
		while (!inStream.eof()) {
			line++;
			inStream.get(c);
			if (inStream.eof()) break;
			switch (c) {
			case 'p':
				inStream.get(c);
				inStream.getline(str, 999, ' ');
				if (strcmp(str, "edge") && strcmp(str, "edges")) {
					exitAndLog("Error reading 'p' line in input file: no 'edge' keyword found. Exiting...\n");
				}
				inStream >> numNodes >> numEdges >> numCols;
				//Set up the 2d adjacency matrix and the precol vector
				adjacent.resize(numNodes, vector<bool>(numNodes, false));
				precol.resize(numNodes, -1);
				colUsed.resize(numCols, false);
				break;
			case 'e':
				int node1, node2;
				inStream >> node1 >> node2;
				if (node1 < 1 || node1 > numNodes || node2 < 1 || node2 > numNodes) {
					exitAndLog("Error. Node number out of range in input file. Exiting...\n");
				}
				node1--;
				node2--;
				if (!adjacent[node1][node2]) {
					edgeCnt++;
				}
				else {
					exitAndLog("Error. Edge defined more thn once in input file. Exiting...\n");
				}
				adjacent[node1][node2] = true;
				adjacent[node2][node1] = true;
				break;
			case 'n':
				inStream >> node >> colour;
				if (node < 1 || node > numNodes || colour < 1 || colour > numCols) {
					exitAndLog("Error. Node or colour out of range in line beginning with 'n' in input file. Exiting...\n");
				}
				node--;
				colour--;
				if (precol[node] != -1) {
					exitAndLog("Error. Precolouring defined more than once in input file. Exiting...\n");
				}
				precol[node] = colour;
				colUsed[colour] = true;
				break;
			case 'c':
				inStream.putback('c');
				inStream.get(str, 999, '\n');
				break;
			default:
				exitAndLog("Error: Unknown line code in input file. Exiting...\n");
			}
			inStream.get(); // Kill the newline;
		}
	}
	catch (...) {
		exitAndLog("Undefined error in input file. It is probably corrupted. Exiting...\n");
	}
	inStream.close();
	if (edgeCnt != numEdges) {
		exitAndLog("Error: Number of edges stated at the top of the input file is not equal to the number found in the file. Exiting...\n");
	}

	//Check that all colours are used in the preassignments
	for (i = 0; i < numCols; i++) {
		if (colUsed[i] == false) {
			exitAndLog("Error: A colour is not being used in the precolourings given in input file. Exiting...\n");
		}
	}

	//If we are here, we assume that a valid input file has been parsed. We now construct the graph G;
	G.n = numNodes;
	G.m = numEdges;
	G.k = numCols;
	G.adj = adjacent;
	G.precol = precol;
	G.precolList.clear();
	for (i = 0; i < G.n; i++) if (G.precol[i] != -1) G.precolList.push_back(i);
	//Finally, we use the adjacency matrix to construct the adjacency list and degree array
	G.deg.resize(numNodes, 0);
	G.adjList.resize(numNodes, vector<int>());
	for (i = 0; i < numNodes; i++) {
		for (j = 0; j < numNodes; j++) {
			if (G.adj[i][j] && i != j) {
				G.adjList[i].push_back(j);
				G.deg[i]++;
			}
		}
	}
	//and calculate the min and max degree
	G.minDeg = G.deg[0];
	G.maxDeg = G.deg[0];
	for (i = 1; i < G.n; i++) {
		if (G.deg[i] > G.maxDeg) G.maxDeg = G.deg[i];
		if (G.deg[i] < G.minDeg) G.minDeg = G.deg[i];
	}
}

void checkSlnValidity(SOL &S) {
	//Does a series of checks to see if a solution is correct. If it isn't, output is reported and the program halts
	bool OK = true;
	int i, j;
	//Check the solution size
	if (G.n != S.col.size()) {
		exitAndLog("Error The graph and the solution do not contain the same number of vertices. Exiting.\n");
	}
	//Check that the claimed cost is valid
	if (S.numHappy < 0 || S.numHappy > G.n) {
		cout << "Error. The claimed cost of " << S.numHappy << " is invalid. (There are " << G.n << " vertices in the graph).\n";
		OK = false;
	}
	//Check it's colours and precolours
	for (i = 0; i < G.n; i++) {
		if (S.col[i] < 0 || S.col[i] >= G.k) {
			cout << "Error. Vertex-" << i + 1 << " contains and invalid colour (" << S.col[i] + 1 << ").\n";
			OK = false;
		}
		if (G.precol[i] != -1) {
			//Vertex i is precoloured in the graph
			if (S.col[i] != G.precol[i]) {
				cout << "Error. Vertex-" << i + 1 << " is assigned to colour-" << S.col[i] + 1 << ", but should actually be preassigned to colour-" << G.precol[i] + 1 << ".\n";
				OK = false;
			}
		}
	}
	//Check that S.items and S.col match
	for (i = 0; i < S.items.size(); i++) {
		for (j = 0; j < S.items[i].size(); j++) {
			if (S.col[S.items[i][j]] != i) {
				cout << "Error: S.items and S.cols do not match.\n";
				OK = false;
			}
		}
	}
	//Now check claimed cost and constency of isHappy array
	int claimedCost = S.numHappy;
	vector<bool> claimedIsHappy = S.isHappy;
	evaluate(S);
	if (claimedCost != S.numHappy) {
		cout << "Error. The claimed score of " << claimedCost << " happy vertices is not equal to the actual score of " << S.numHappy << ".\n";
		OK = false;
	}
	for (i = 0; i < G.n; i++) {
		if (claimedIsHappy[i] != S.isHappy[i]) {
			cout << "Error. The claimed happyness of vertex " << i + 1 << " is not equal to the actual happyness of " << S.isHappy[i] << ".\n";
			OK = false;
		}
	}
	if (!OK) {
		exitAndLog("Problem instance and solutions do not match. Exiting.\n");
	}
}