#include "initsol.h"

extern GRAPH G;

void emptyColourClass(SOL &S, int c) {
	//Removes all free vertices from colour class c
	int i, v;
	int numToDelete = S.items[c].size() - S.numPreass[c];
	for (i = 0; i < numToDelete; i++) {
		v = S.items[c].back();
		S.col[v] = -1;
		S.items[c].pop_back();
	}
}

void fillColourClass(SOL &S, int c) {
	//Puts all free vertices into colour class c. Assumes S.items only contains precoloured at this point
	int v;
	for (v = 0; v < G.n; v++) {
		if (G.precol[v] == -1) {
			S.items[c].push_back(v);
			S.col[v] = c;
		}
	}
}

void GreedyMHV(SOL &S) {
	int j, bestJ, maxHappy = INT_MIN;
	for (j = 0; j < G.k; j++) {
		//Assign all free vertices to colour class j
		fillColourClass(S, j);
		evaluate(S);
		//Check if this is the best so far
		if (S.numHappy > maxHappy) {
			maxHappy = S.numHappy;
			bestJ = j;
		}
		emptyColourClass(S, j);
	}
	//Produce the solution by putting all free vertices into the best colour class
	fillColourClass(S, bestJ);
	evaluate(S);
}


//-------------------------------------------------
int determineColouredVertexStatus(int v, SOL &S, vector<int> &status, int &numHappy) {
	//Determines the status of a coloured vertex v by examining it neighbours u. Updates the numHappy variable where apt.
	if (S.col[v] < 0) exitAndLog("Error: have called the determineColouredVertexStatus function with an uncoloured vertex. Exiting...\n");
	//If a coloured vertex is already happy or destined to be unhappy, its status cannot change, so we end
	if (status[v] == 0 || status[v] == 2) return status[v];
	
	//The status of this vertex is P and therfore has the potential to change, so calculate its new status
	int oldstatus = status[v];
	int i, u;
	bool freeNeighbour = false;
	for (i = 0; i < G.adjList[v].size(); i++) {
		u = G.adjList[v][i];
		if (S.col[u] == -1) {
			freeNeighbour = true;
		}
		else if (S.col[v] != S.col[u]) {
			//v cannot be happy, so end immediately
			return 2;
		}
	}
	//If we are here, no neighbours of v are coloured differently
	if (!freeNeighbour) {
		//v has become happy
		numHappy++;
		S.isHappy[v] = true;
		return 0;
	}
	else {
		//v has uncoloured neighbours and has the potential to be happy
		return 1;
	}
}

int determineFreeVertexStatus(int v, SOL &S, vector<int> &status) {
	//Determines the status of an uncoloured (free) vertex v by examining its neighbours u
	if (S.col[v] >= 0) exitAndLog("Error: have called the determineFreeVertexStatus function with a coloured vertex. Exiting...\n");
	int i, u, anAdjColour = -1;
	bool multipleAdjColoursToV = false;
	for (i = 0; i < G.adjList[v].size(); i++) {
		u = G.adjList[v][i];
		if (status[u] == 1) {
			//v is an L_p vertex (adjacent to a potentially happy, coloured vertex u). End immediately
			return 3;
		}
		if (S.col[u] != -1 && anAdjColour == -1) {
			//at least one colour adjacent to v
			anAdjColour = S.col[u];
		}
		else if (S.col[u] != -1 && S.col[u] != anAdjColour) {
			//have observed at least two diff colours adjacent to v, so v cannot be happy
			multipleAdjColoursToV = true;
		}
	}
	//If we are here, v is not adjacent to a P-vertex -- it's status must be 4, 5, or 6.
	if (anAdjColour != -1 && !multipleAdjColoursToV) {
		//v is adjacent to vertices of only one colour, it has the potential to be happy
		return 4;
	}
	else if (multipleAdjColoursToV) {
		//v is adjacent to vertices of more than one colour, it is destined to be unhappy
		return 5;
	}
	else {
		//all of v's neighbours are uncoloured.
		return 6;
	}
}

int getNextVertexToColour(vector<int> &status, bool &PVertexExists, bool &LHVertexExists, bool &LUVertexExists) {
	int i;
	PVertexExists = LHVertexExists = LUVertexExists = false;
	int LUVertexPos = -1, LFVertexPos = -1;
	//Look for a coloured, potentially happy vertex v (a P-vertex)
	for (i = 0; i < G.n; i++) {
		if (status[i] == 1) {
			PVertexExists = true;
			return i;
		}
	}
	//If we are here, we need to look for an uncoloured potentially happy, and then uncoloured unhappy, then uncoloured free vertex
	for (i = 0; i < G.n; i++) {
		if (status[i] == 4) {
			LHVertexExists = true;
			return i;
		}
		else if (status[i] == 5) {
			LUVertexExists = true;
			LUVertexPos = i;
		}
		else if (status[i] == 6) {
			LFVertexPos = i;
		}
	}
	//If we are here, we have not found a P-vertex or a LH-vertex, but we have found a LU-vertex or, failing that, a LF-vertex
	if (LUVertexExists) {
		return LUVertexPos;
	}
	else if (LFVertexPos < status.size()) {
		return LFVertexPos;
	}
	else {
		exitAndLog("Error: Should not be here in the getNextVertexToColour function\n");
		return 0;
	}
}

void updateNeighboursStatus(int v, SOL &S, vector<int> &status, int &numHappy) {
	//A vertex v has just been coloured. Update the status of all neigbours
	int u, j;
	for (j = 0; j < G.adjList[v].size(); j++) {
		u = G.adjList[v][j];
		if (S.col[u] != -1) {
			status[u] = determineColouredVertexStatus(u, S, status, numHappy);
		}
		else {
			status[u] = determineFreeVertexStatus(u, S, status);
		}
	}
}

void updateDistanceTwo(int v, vector<int> &status, SOL &S, int &numHappy) {
	//Have had to modernise this function from Li and Zhang's algorithm description. Imagine we have a chain of
	//vertices as follows [(1: P, red), (2: LP), (3: P, blue), (4: LP)] and we select vertex-1. This causes v-2 to become red too. Hence v-1 is happy, and v-2 is unhappy.
	//As a result v-3 now also changes it's state to unhappy, which should also change the status of v-4 to either LH or LU.
	//However, this will not happen with the Li-Zhang algorithm, because it tells us only to update statuses of distance 2 from v. So if we have a vertex of distance-2 from 
	//v that is coloured and had a status P (i.e. 1) that has now changed, we should also check and update its uncoloured neighbours statuses.
	
	//A vertex v and its neighbours have just been coloured with the same colour and v is happy. Update the statuses of all vertices of distance 2 or less from v
	int u, w, x, j, l, m, oldState;
	status[v] = 0;
	numHappy++;
	S.isHappy[v] = true;
	for (j = 0; j < G.adjList[v].size(); j++) {
		u = G.adjList[v][j];
		status[u] = determineColouredVertexStatus(u, S, status, numHappy);
	}
	//update status of each vertex w adjacent to the neighbours u of v
	for (j = 0; j < G.adjList[v].size(); j++) {
		u = G.adjList[v][j];
		for (l = 0; l < G.adjList[u].size(); l++) {
			w = G.adjList[u][l];
			if (S.col[w] != -1) {
				oldState = status[w];
				status[w] = determineColouredVertexStatus(w, S, status, numHappy);
				//This is the additional bit
				if (oldState != status[w]) {
					for (m = 0; m < G.adjList[w].size(); m++) {
						x = G.adjList[w][m];
						if (S.col[x] == -1) {
							status[x] = determineFreeVertexStatus(x, S, status);
						}
					}
				}
				//
			}
			else {
				status[w] = determineFreeVertexStatus(w, S, status);
			}
		}
	}
}

int getAdjacentColour(int v, SOL &S) {
	//Return a colour adjacent to vertex v
	int j, u;
	for (j = 0; j < G.adjList[v].size(); j++) {
		u = G.adjList[v][j];
		if (S.col[u] != -1) {
			return S.col[u];
		}
	}
	//If we are here there is a problem
	exitAndLog("Error: should not be here in function getAdjacentColour\n");
	return 0;
}

int getUnhappyColouredNeighbour(int v, vector<int> &status) {
	//Return a vertex adjacent to v that is unhappy
	int j, u;
	for (j = 0; j < G.adjList[v].size(); j++) {
		u = G.adjList[v][j];
		if (status[u] == 2) {
			return u;
		}
	}
	//If we are here there is a problem
	exitAndLog("Error: should not be here in function getUnhappyColouredNeighbour\n");
	return 0;
}

void GrowthMHV(SOL &S) {
	//GROWTH-MHV algorithm. Labels in the status array are as follows:
	// 0: (H-Vertex) Coloured and happy
	// 1: (P-vertex) Coloured and has potential to be happy
	// 2: (U-vertex) Coloured and destined to be unhappy
	//---------
	// 3: (L_p-vertex) Not yet coloured, and adjacent to a P-vertex 
	// 4: (L_h-vertex) Not yet coloured, not adjacent to a P-vertex, but has potential to be happy 
	// 5: (L_u-vertex) Not yet coloured, not adjacent to a P-vertex, but destined to be unhappy 
	// 6: (L_f-vertex) Not yet coloured, and not adjacent to any colured vertex
	int v, u, i, j;
	int numColoured = G.precolList.size();
	int numHappy = 0;

	//Determine initial status of all coloured vertices, and then all uncoloured vertices
	vector<int> status(G.n, -1);
	for (i = 0; i < G.n; i++) {
		if (S.col[i] != -1) {
			status[i] = determineColouredVertexStatus(i, S, status, numHappy);
		}
	}
	for (i = 0; i < G.n; i++) {
		if (S.col[i] == -1) status[i] = determineFreeVertexStatus(i, S, status);
	}

	//Main algorithm
	bool PVertex, LHVertex, LUVertex;
	while (numColoured < G.n) {
		//Chooose next vertex v to colour 
		v = getNextVertexToColour(status, PVertex, LHVertex, LUVertex);
		if (PVertex) {
			//v is a P-vertex with colour i. Colour all of its uncoloured neighbours u, with colour i
			i = S.col[v];
			for (j = 0; j < G.adjList[v].size(); j++) {
				u = G.adjList[v][j];
				if (status[u] == 3 && S.col[u] == -1) {
					S.col[u] = i;
					S.items[i].push_back(u);
					numColoured++;
				}
			}
			//update status of all vertices within distance 2 of v
			updateDistanceTwo(v, status, S, numHappy);
		}
		else if (LHVertex) {
			//v is a LH-vertex adjacent to a colour i. Colour v and all its its uncoloured neighbours u with i.
			i = getAdjacentColour(v, S);
			S.col[v] = i;
			S.items[i].push_back(v);
			numColoured++;
			for (j = 0; j < G.adjList[v].size(); j++) {
				u = G.adjList[v][j];
				if (S.col[u] == -1) {
					S.col[u] = i;
					S.items[i].push_back(u);
					numColoured++;
				}
			}
			//update status of all vertices within distance 2 of v
			updateDistanceTwo(v, status, S, numHappy);
		}
		else if (LUVertex) {
			//v is an LU-vertex. When we colour it, it will be unhappy. Also, it must have an unhappy neighbour currently
			u = getUnhappyColouredNeighbour(v, status);
			i = S.col[u];
			S.col[v] = i;
			S.items[i].push_back(v);
			numColoured++;
			status[v] = 2;
			//update the status of v's neighbours
			updateNeighboursStatus(v, S, status, numHappy);
		}
		else {
			//v must be an LF-vertex (i.e. v is in a component in which none of the vertices are coloured). First choose a random colour for v
			i = rand() % G.k;
			S.col[v] = i;
			S.items[i].push_back(v);
			numColoured++;
			if (G.adjList[v].empty()) {
				//v is an isolated vertex so it is happy
				status[v] = 0;
				S.isHappy[v] = true;
				numHappy++;
			}
			else {
				//v has neighbours so we set it as potentially happy and update its neighbours
				status[v] = 1;
				updateNeighboursStatus(v, S, status, numHappy);
			}
		}
	}
	S.numHappy = numHappy;
}

void makeInitSol(SOL &S, int algorithm) {
	int i, v;
	S.items.resize(G.k, vector<int>());
	S.numPreass.resize(G.k, 0);
	S.col.resize(G.n, -1);
	S.isHappy.resize(G.n, false);
	S.numHappy = 0;

	//Assign precolured nodes to colours classes
	for (i = 0; i < G.precolList.size(); i++) {
		v = G.precolList[i];
		S.items[G.precol[v]].push_back(v);
		S.numPreass[G.precol[v]]++;
		S.col[v] = G.precol[v];
	}

	//Now produce the solution
	if (algorithm == 1) {
		//Assign all free vetices to each colour class in turn, evaluate, and take the best of these k solutions
		GreedyMHV(S);
	}
	else {
		//Produce a solution using the Li-Zhang algorithm
		GrowthMHV(S);
	}
}

